/* 1450, Tue 16 Sep 02 (CEST)

   FD_FILTER.C:  Performs tidy-up operations on NeTraMet flow-data files
      * Compute flow rates [To/From ByteRate, To/From PduRate]
      * Change (NeTraMet) format for file
      * Filter out flows from file, tag those remaining [FlowTag]
      * Filter out #Stats records

   Copyright (C) 1994-2002 by Russell Fulton & Nevil Brownlee
   University of Auckland | CAIDA */

/*
 * $Log: fd_filter.c,v $
 * Revision 1.1.1.2.2.17  2002/10/31 23:02:26  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 *
 * Revision 1.1.1.2.2.12  2002/02/23 01:57:16  nevil
 * Moving srl   amples to examples/ directory.  Modified examples/Makefile.in
 *
 * Revision 1.1.1.2.2.7  2000/11/29 21:40:08  nevil
 * Bug fixes for stream / packet queue handling
 *
 * Revision 1.1.1.2.2.6  2000/08/08 19:44:42  nevil
 * 44b8 release
 *
 * Revision 1.1.1.2.2.4  2000/06/06 03:38:08  nevil
 * Combine NEW_ATR with TCP_ATR, various bug fixes
 *
 * Revision 1.1.1.2.2.1  2000/01/12 02:57:02  nevil
 * Implement 'packet pair matched' turnaroundtime distribution attributes.
 * Fix ASN-related bugs in NeTraMet, distribution-related bugs in fd_filter.
 *
 * Revision 1.1.1.2  1999/10/03 21:06:14  nevil
 * *** empty log message ***
 *
 * Revision 1.1.1.1.2.9  1999/09/22 05:38:33  nevil
 * Improve code to work properly on 64-bit machines
 * - Add OS=ALPHA handling to configure.in
 * - Clean up the Alpha compiler warnings
 * - Change all the snmp-related code to use Bit32 instead of unsigned long
 *
 * Revision 1.1.1.1.2.8  1999/09/14 00:46:45  nevil
 * 4.3 Release ..
 *  - Implement -D option to run NeMaC as a daemon
 *  - Tidy up the on-line help displays
 *  - Make IPv4 the default PeerType for sfmt_attribute()
 *
 * Revision 1.1.1.1.2.7  1999/05/18 03:36:24  nevil
 * Implement IPv6 in NeTraMet, and its manager/collectors.
 * - This is controlled by the V6 #define
 * - NeTraMet recognises v6 packets and fishes through their extension
 *     headers until it finds the actual payload.
 * - NeMaC et al display v6 addresses in the fom specified by RFC 2373
 * - fd_util and fd_extract allow colons in addresses (for defining tags)
 *
 * Revision 1.1.1.1.2.6  1999/02/15 21:24:05  nevil
 * Distribution file for 4.3b9
 *
 * Revision 1.1.1.1.2.5  1999/02/03 04:41:41  nevil
 * Implementation of TCP attributes, part 5
 *
 * Revision 1.1.1.1.2.4  1999/01/27 04:26:12  nevil
 * Minor corrections to fix compiler warnings
 *
 * Revision 1.1.1.1.2.3  1999/01/20 04:01:32  nevil
 * Implementation of TCP attributes, part 4
 *
 * Revision 1.1.1.1.2.2  1999/01/08 01:37:02  nevil
 * Implement TCPdata and DScodepoint attribute
 *
 * Revision 1.1.1.1.2.1  1998/11/26 04:13:49  nevil
 * Fix problems in fd_filter and fd_extract when dealing with 'new'
 * attributes, especially those used with NETFLOW. In particular:
 * - Make sure attribs[] is referred to via attr_ix[]
 * - Get the separators correctly from a #Format record (no quotes)
 * - fd_extract wasn't writing out its data for the last sample!
 *
 * Revision 1.1.1.1  1998/11/16 03:57:26  nevil
 * Import of NeTraMet 4.3b3
 *
 * Revision 1.1.1.1  1998/11/16 03:22:00  nevil
 * Import of release 4.3b3
 *
 * Revision 1.1.1.1.2.1  1998/11/11 23:14:36  nevil
 * Only include malloc.h if we HAVE_MALLOC_H
 *
 * Revision 1.1.1.1  1998/10/28 20:31:23  nevil
 * Import of NeTraMet 4.3b1
 *
 * Revision 1.1.3.2.2.1  1998/10/27 04:39:13  nevil
 * 4.3b1 release
 *
 * Revision 1.1.3.2  1998/10/18 23:44:04  nevil
 * Added Nicolai's patches, some 'tidying up' of the source
 *
 * Revision 1.1.3.1  1998/10/13 02:48:16  nevil
 * Import of Nicolai's 4.2.2
 *
 * Revision 1.1.1.1  1998/08/24 12:09:28  nguba
 * NetraMet 4.2 Original Distribution
 *
 * Revision 1.3  1998/05/07 04:28:53  rtfm
 * Implement NetFlowMet, the Cisco NetFlow RTFM meter
 */

#if HAVE_CONFIG_H
#include <ntm_conf.h>
#endif

#define DEBUG       0
#define TEST_DELTA  0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if HAVE_MALLOC_H
# include <malloc.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>


/*
   filter fmt_file [ in_file [ trailer_file [ > out_file ]]]
*/

#include "ausnmp.h"
#include "asn1.h"

#include "fd_data.h"
#include "nmc_c64.h"

#define OK        0
#define NODATA    1 
#define NOFILE    2
#define CANTOPEN  3

int status;
char *stats[] = {
   "OK", "No Data", "No File", "Can't Open input"
   };

int max_flows = 0;
int recno =  0;

int main(int argc, char *argv[])
{
   int a, tail_file = 0, utc;
   struct flow_stats in_set, out_set;

   char format_fn[STR_LEN], input_fn[STR_LEN], trailer_fn[STR_LEN];

   int read_to = 900, warn_freq = 3600;
   time_t t, last_read = time(0), dead_warn;
   struct stat statbuf, statbuf_o;

   if (argc < 2) {
      fprintf(stderr, 
         "%s format_file [ input_file [ trailer_file [ > output_file ]]]\n",
         argv[0]);
      exit(0);
      }

   memset(&in_set, 0, sizeof(struct flow_stats));
   memset(&out_set, 0, sizeof(struct flow_stats));
   init_attribs();

   format_fn[0] = input_fn[0] = trailer_fn[0] = '\0';
   quiet = 0;
   listrules = 1;  /* List tags and format (from flow data file) */
   listformat = 0;  /* From the format file */

   for (a = 1; a != argc; ++a) {
      if (argv[a][0] == '-') {
	 switch (argv[a][1]) {
	 case 'f':
	    listformat = 1;
	    break;
	 case 'q':
	    quiet = 1;  /* No 'comment' listing */
	    break;
	 case 't':
	    tail_file = 1;
	    break;
	 default:
	    fprintf(stderr,"Invalid option: -%c\n", argv[a][1]);
	    exit(0);
	    }
         continue;
	 }
      if (format_fn[0] == '\0') {
	 strncpy(format_fn, argv[a], STR_LEN);
	 format_fn[STR_LEN-1] = '\0';
         }
      else if (input_fn[0] == '\0') {
	 strncpy(input_fn, argv[a], STR_LEN);
	 input_fn[STR_LEN-1] = '\0';
         }
      else if (trailer_fn[0] == '\0') {
	 strncpy(trailer_fn, argv[a], STR_LEN);
	 trailer_fn[STR_LEN-1] = '\0';
         }
      }

   if (format_fn[0] == '\0') {
      fprintf(stderr,"No format file specified!\n");
      exit(1);
      }

   if (trailer_fn[0] != '\0' && tail_file) {
      fprintf(stderr,"can not have both trailer file and tail_file\n");
      exit(1);
      }

   parse_format(format_fn, &out_set);

   if (listrules && !quiet)
      list_tags();

   if (argc <= 2) in_set.flows = stdin;

   out_set.flows = stdout;  /* Alway write to stdout */

   recno = max_flows = 0;  

   /* Outer loop to enable fd_filter to tail the output of an 
      active flow file -- controlled by tail_file (-t) */
   quack(123);
   for ( ; ; sleep(60)) {
      t = time(0);

      if (t - last_read > read_to && t - dead_warn > warn_freq) {
         fprintf( out_set.flows,
            "#Warning: No flow data for %d seconds: %s\n",
            t - last_read, stats[status]);
         dead_warn = t;
         }

      if (in_set.flows != NULL) fstat(fileno(in_set.flows), &statbuf_o);

      if (stat(input_fn, &statbuf ) != 0  &&      /* Failed */
            in_set.flows != NULL &&               /* Have file open */
            last_read > statbuf_o.st_mtime ) {    /* Old one unchanged */     
         status = NOFILE;
         continue;
         }

      if (in_set.flows == NULL ||                   /* First time */
            (statbuf.st_ino != statbuf_o.st_ino &&  /* New file and  */
	    last_read > statbuf_o.st_mtime))  {     /* Old one unchanged */
	 int save_flow_no = max_flows;
	 /* Close and reopen input file */
	
         if (in_set.flows != NULL) fclose(in_set.flows);
	 if ((in_set.flows = fopen(input_fn, "r")) == NULL) {
            if (tail_file) {
	       status = CANTOPEN;
	       continue;
	       }
	    else {
	       fprintf(stderr, "Can't open input %s:",input_fn);
	       exit(1);
	       }
	    } 

         /* Read the header line and check it */
         max_flows = process_header(&utc, &in_set, &out_set); 
	    
         if (save_flow_no < max_flows) {
	    /* First or meter has been restated with more flows */ 
	    if (save_flow_no) {  /* Not first time */
               free_flow_mem(max_flows);
               }

	    alloc_flows(&in_set, &out_set, max_flows);
	    read_format_rec(&in_set, &out_set);
            write_filter_info(argc,argv, utc, &out_set);
#if NEW_ATR
	    init_dists(&out_set);
#endif
            }
         }  /* Opening files */
      status = NODATA;

      while (fgets(inbuf, sizeof(inbuf), in_set.flows) != NULL) { 
	 recno++;
	 process_rec(inbuf, &in_set, &out_set);
	 last_read = time(NULL);
	 status = OK;
         }
      if (!tail_file) break;

      fflush(out_set.flows);   /* Flush each batch */
      fseek(in_set.flows, 0L, SEEK_CUR);  /* Reset eof */
      }
   fclose(in_set.flows);

   /* Now handle trailer file if required, and process
      just one sample of data */

   if (trailer_fn[0] != '\0') {
      int Data_seen = 0;

      if (!open_trailer(trailer_fn, &in_set)) exit(1);
      while (fgets(inbuf, sizeof(inbuf), in_set.flows) != NULL) {
         recno++;
         if (process_rec(inbuf, &in_set, &out_set) == DATA_REC){
            Data_seen = 1;
            }
         else { if (Data_seen) break; }
         }

     fclose(in_set.flows);
     }

   fclose(out_set.flows);
   }


int open_trailer(char *trailer_fn,
   struct flow_stats *in_set, struct flow_stats *out_set)
{
   if (trailer_fn[0] == '\0') return 0;  /* No trailer file */
   if ((in_set->flows=fopen(trailer_fn, "r")) == NULL) {
      fprintf(stderr, "can't open trailer %s:",trailer_fn); perror("");
      return 0;
      }
      
   recno = 0;
   if (fgets(inbuf, sizeof(inbuf),in_set->flows ) == NULL)
        /* Read header record */
      return 0;
   if (strncmp(inbuf, "##NeTraMet ", 10) != 0) {
      fprintf(stderr,"Trailer not a NeTraMet file!\n");
      return 0;
      }

   for (;;) {
      ++recno;
      if (fgets(inbuf, sizeof(inbuf),in_set->flows ) == NULL) {
         fprintf(stderr,"Failed to find format record in trailer file\n");
         return 0;
         }
      if (inbuf[0] != '#') {
         fprintf(stderr,"Non-# record found in trailer file header\n");
         return 0;;
         }
      if (inbuf[1] == 'F')  /* Should be #Format */
         break;
      }
   ic = ' ';  ibp = inbuf+8;  /* Scan format line */
   scan_format(in_set, 1, listformat);  /* Single_record */
   if (!check_formats(in_set, out_set)) return 0;
   ++recno;
   }
